/* icmplog.c of iplog v1.5 -- eric@ojnk.org */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <arpa/inet.h>
#include <time.h>
#include <string.h>
#include <netdb.h>
#define __BSD_SOURCE
#define __FAVOR_BSD
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include "include/icmp-compat.h"
#include "include/iplog.h"
#include "include/lockfile.h"

static char *protolookup(int);
static char *print_time(u_int);
static char *my_ntoa(struct in_addr);
static char *return_duration(u_int);

struct packet {
	struct ip ip;
	struct icmp icmp;
	struct tcphdr tcp;
};

static void print_routers(struct packet);

#ifdef DETECT_SMURF
#include "include/smurf.h"
#endif

#define MIN(a,b)	((a < b) ? (a) : (b))
#define ICMP_SIZE	sizeof(struct ip)

#define proto	pkt.icmp.icmp_ip.ip_p
#define gw		pkt.icmp.icmp_gwaddr
#define src		pkt.ip.ip_src
#define dest	pkt.icmp.icmp_ip.ip_dst
#define orig	pkt.icmp.icmp_ip.ip_src

static const char *codes[] =
{
	"echo reply from %s sz=%u(+%u)",	/* ICMP_ECHO_REPLY */
	"undefined icmp type 1 recieved from %s(?)",	/* ? */
	"undefined icmp type 2 recieved from %s(?)",	/* ? */
	"destination unreachable from %s",	/* ICMP_DEST_UNREACHABLE */
	"source quench from %s",	/* ICMP_SOURCE_QUENCH */
	"redirect message from %s",	/* ICMP_REDIRECT */
	"undefined icmp type 6 recieved from %s(?)",	/* ? */
	"undefined icmp type 7 recieved from %s(?)",	/* ? */
	"echo from %s sz=%u(+%u)",	/* ICMP_ECHO */
	"router advertisement from %s",		/* ICMP_ROUTER_ADVERT (rfc1256) */
	"router solicitation from %s",	/* ICMP_ROUTER_SOLICIT (rfc1256) */
	"time exceeded (%s)",		/* ICMP_TIME_EXCEEDED */
	"parameter problem from %s (ptr %d)",	/* ICMP_PARAMETER_PROBLEM */
	"timestamp request from %s",	/* ICMP_TIMESTAMP */
	"timestamp reply from %s (%s)",		/* ICMP_TIMESTAMP_REPLY */
	"info request from %s",		/* ICMP_INFO_REQUEST */
	"info reply from %s",		/* ICMP_INFO_REPLY */
	"address mask request from %s",		/* ICMP_ADDRESS */
	"address mask reply from %s (%s)"	/* ICMP_ADDRESS_REPLY */
};

static const char *unreach[] =
{
	"%s: network is unreachable [%s: dp=%u sp=%u]",		/* ICMP_NET_UNREACH */
	"%s: host is unreachable [%s: dp=%u sp=%u]",	/* ICMP_HOST_UNREACH */
	"%s: %s: protocol is unreachable",	/* ICMP_PROT_UNREACH */
	"%s: %s port is unreachable [dp=%u sp=%u]",		/* ICMP_PORT_UNREACH */
	"%s: fragmentation needed (IP_DF set)",		/* ICMP_FRAG_NEEDED */
	"%s: source route failed [%s: dp=%u sp=%u]",	/* ICMP_SR_FAILED */
	"%s: network is unreachable(?) [%s: dp=%u sp=%u]",	/* ICMP_NET_UNKNOWN */
	"%s: host is unreachable(?) [%s: dp=%u sp=%u]",		/* ICMP_HOST_UNKNOWN */
	"%s: host is isolated from source",		/* ICMP_HOST_ISOLATED */
	"%s: network is unreachable (admin prohibited) ",	/* ICMP_NET_ANO */
	"%s: host is unreachable (admin prohibited)",	/* ICMP_HOST_ANO */
	"%s: network is unreachable (tos is prohibited)",	/* ICMP_NET_UNR_TOS */
	"%s: host is unreachable (tos is prohibited)",	/* ICMP_HOST_UNR_TOS */
	"%s: packet filtered [%s: dp=%u sp=%u]",	/* ICMP_PKT_FILTERED */
	"%s: precedence violation",	/* ICMP_PREC_VIOLATION */
	"%s: precedence cutoff"		/* ICMP_PREC_CUTOFF */
};

static const char *redir[] =
{
	"(%s) redirect %s to network %s",	/* ICMP_REDIR_NET */
	"(%s) redirect %s to host %s",	/* ICMP_REDIR_HOST */
	"(%s) redirect-tos %s to network %s",	/* ICMP_REDIR_NETTOS */
	"(%s) redirect-tos %s to host %s"	/* ICMP_REDIR_HOSTTOS */
};

int main(void)
{
	int sock;
	struct in_addr in;
	struct packet pkt;

#ifdef DETECT_SMURF
	signal(SIGALRM, sigalrm);
#endif

	if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		perror("socket");
		exit(-1);
	}
	openlog("icmplog", LOGGING_PRIORITY, LOGGING_FACILITY);
	signal(SIGHUP, SIG_IGN);

	fork_to_background();
	initialize_lockfile("icmplog");
	drop_privileges();

	for (;;) {
		read(sock, (struct packet *)&pkt, sizeof(pkt));

		if (pkt.ip.ip_hl != 5) {
			syslog(LEVEL, "suspicious ip options from %s", inet_ntoa(src));
			continue;
		}
		switch (MIN(pkt.icmp.icmp_type, 19)) {
			case ICMP_TIME_EXCEEDED:
				break;			/* ignored */
			case ICMP_ECHO_REPLY:
#ifdef DETECT_SMURF
				if (smurf(pkt))
					syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), UNFIX(pkt.ip.ip_len) - ICMP_SIZE, ICMP_SIZE);
				break;
#endif
			case ICMP_ECHO:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), UNFIX(pkt.ip.ip_len) - ICMP_SIZE, ICMP_SIZE);
				break;
			case ICMP_DEST_UNREACHABLE:
				if (pkt.icmp.icmp_code > UNREACHABLE_MAX) {
#ifdef LOG_WIERD_ICMP
					syslog(LEVEL, "undefined destination unreachable code from %s(?)", inet_ntoa(src));
#endif
					break;
				}
				syslog(LEVEL, unreach[(int)pkt.icmp.icmp_code], inet_ntoa(src), protolookup(proto), UNFIX(pkt.tcp.th_dport), UNFIX(pkt.tcp.th_sport));
				break;
			case ICMP_REDIRECT:
				if (pkt.icmp.icmp_code > REDIRECT_MAX) {
#ifdef LOG_WRITE_ICMP
					syslog(LEVEL, "undefined redirect code from %s(?)", inet_ntoa(src));
#endif
					break;
				}
				if (gw.s_addr != src.s_addr)
					syslog(LEVEL, "warning: gateway address (%s) != source address (%s)", my_ntoa(gw), my_ntoa(src));
				syslog(LEVEL, redir[(int)pkt.icmp.icmp_code], my_ntoa(gw), my_ntoa(orig), my_ntoa(dest));
				break;
			case ICMP_TIMESTAMP_REPLY:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), print_time(pkt.icmp.icmp_ttime));
				break;
			case ICMP_ADDRESS_REPLY:
				in.s_addr = htonl(pkt.icmp.icmp_mask);
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], my_ntoa(src), my_ntoa(in));
				break;
			case ICMP_PARAMETER_PROBLEM:
				syslog(LEVEL, codes[(int)pkt.icmp.icmp_type], inet_ntoa(src), pkt.icmp.icmp_pptr);
				break;
			case ICMP_ROUTER_ADVERT:
				print_routers(pkt);
				break;
			case ICMP_UNDEFINED:
#ifdef LOG_WEIRD_ICMP
				syslog(LEVEL, "undefined icmp type %d from %s(?)", (int)pkt.icmp.icmp_type, inet_ntoa(src));
#endif
				break;
			default:
				syslog(LEVEL, codes[(int)MIN(pkt.icmp.icmp_type, 19)], inet_ntoa(src));
				break;
		}
	}
}

static char *protolookup(int num)
{
	struct protoent *pe;

	if ((pe = getprotobynumber(num)) != NULL)
		return (pe->p_name);
	else {
		static char buff[16];

		sprintf(buff, "proto %d", num);
		return (buff);
	}
}

/* inet_ntoa() uses a statically allocated variable, gotta love that */
static char *my_ntoa(struct in_addr in)
{
	char host[32];
	unsigned char *ip;

	ip = (unsigned char *)&in;
	sprintf(host, "%d.%d.%d.%d", (ip[0] & 255), (ip[1] & 255), (ip[2] & 255), (ip[3] & 255));

	return (strdup(host));
}


static char *return_duration(u_int time)
{
	static char ret[32];

	if (time < 60)
		sprintf(ret, "%u sec", time);
	else
		sprintf(ret, "%u min, %u sec", (time / 60), (time % 60));

	return (strdup(ret));
}

static void print_routers(struct packet pkt)
{
	struct id_rdiscovery *routers;
	char msg[768], temp[64];
	u_char n;

	memset(msg, 0, sizeof(msg));

	routers = (struct id_rdiscovery *)&pkt.icmp.icmp_dun.id_rdiscovery;

	for (n = 1; n <= pkt.icmp.icmp_num_addr; n++, routers++) {
		memset(temp, 0, sizeof(temp));
		snprintf(temp, 64, "%s%s", (n > 1) ? " -> " : "", inet_ntoa(routers->router_addr));
		snprintf(msg, 768, "%s%s", msg, temp);
	}

	syslog(LEVEL, "path of router advertisement from %s: %s (lifetime %s)", my_ntoa(src), msg, return_duration(UNFIX(pkt.icmp.icmp_lifetime)));
	return;
}

static char *print_time(u_int time)
{
	struct tm *tm;
	char ret[32];
	u_long actual_time;

	actual_time = ntohl(time) / 1000;
	tm = localtime((time_t *) actual_time);
	sprintf(ret, "%.2d:%.02d:%.02d", tm->tm_hour, tm->tm_min, tm->tm_sec);

	return (strdup(ret));
}
