/* iplog.h of iplog v1.5 -- eric@ojnk.org */
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <pwd.h>
#include <grp.h>
#include <fcntl.h>
#include <errno.h>
#include <endian.h>
extern int errno;

#ifndef LOGFILE
#include <syslog.h>
#else
void openlog(const char *, int, int);
void syslog(int, const char *fmt,...);

#endif

#ifndef __linux__
#error "iplog compiles on Linux -ONLY-"
#endif

#include "../config.h"

#ifndef DROP_USER
#define DROP_USER "nobody"
#endif

#ifndef DROP_GROUP
#define DROP_GROUP "nogroup"
#endif

#ifndef LOGGING_OPTIONS
#define LOGGING_OPTIONS	LOG_PID
#endif

#ifndef LOGGING_FACILITY
#define LOGGING_FACILITY LOG_DAEMON
#endif

#ifndef LOGGING_PRIORITY
#define LOGGING_PRIORITY	LOG_NOTICE
#endif

#define LEVEL	LOGGING_PRIORITY

#define FIX(x)		network_byte_ordering(x)
#define UNFIX(x)	network_byte_ordering(x)

u_int16_t network_byte_ordering(u_int16_t x)
{
#if BYTE_ORDER == LITTLE_ENDIAN
	u_char *s = (u_char *) & x;

	return (u_int16_t) (s[0] << 8 | s[1]);
#else
	return x;
#endif
}

void fatal(const char *fmt,...)
{
	char *p = malloc(512);
	va_list ap;

	if (p == NULL)
		exit(-1);

	va_start(ap, fmt);
	(void)vsnprintf(p, 512, fmt, ap);
	va_end(p);

	syslog(LOG_WARNING, p);
	free(p);

	exit(-1);
}


void fork_to_background(void)
{
	switch (fork()) {
		case -1:
			_exit(0);
		case 0:
			break;
		default:
			_exit(0);
	}

	if (setsid() == -1)
		fatal("setsid: %m");

	chdir("/");
	umask(022);

	(void)close(fileno(stdin));
	(void)close(fileno(stdout));
	(void)close(fileno(stderr));

	return;
}

void assert_id(uid_t uid, gid_t gid)
{
	struct passwd *pw;
	struct group *gr;
	uid_t my_uid;
	gid_t my_gid;

	if ((pw = getpwnam(DROP_USER)) == NULL) {
		if ((pw = getpwnam("nobody")) != NULL) {
			my_uid = pw->pw_uid;
		} else {
			my_uid = 65534;
		}
	} else
		my_uid = pw->pw_uid;

	if ((gr = getgrnam(DROP_GROUP)) != NULL)
		my_gid = gr->gr_gid;
	else if (((gr = getgrnam("nobody")) != NULL) || ((gr = getgrnam("nogroup")) != NULL))
		my_gid = gr->gr_gid;
	else if (pw != NULL)
		my_gid = pw->pw_gid;
	else
		my_gid = 65534;

	uid = my_uid;
	gid = my_gid;

	return;
}

void drop_privileges(void)
{
	gid_t gid = 65534;
	uid_t uid = 65534;

	assert_id(uid, gid);

	if (setregid(gid, gid) < 0)
		fatal("setregid: %m");

	if (setgroups(1, &gid) < 0)
		fatal("setgroups: %m");

	if (setreuid(uid, uid) < 0)
		fatal("setreuid: %m");

	return;
}

#ifdef LOGFILE
int log_fd, times = 0;
char *last_message = NULL;
time_t last_time = 0L;

void openlog(const char *name, int unused, int ununsed)
{
	char *log_name;

	last_message = strdup("x");

	log_name = malloc(strlen(LOGFILE) + strlen(name) + 1);
	if (log_name == NULL)
		fatal("openlog: %m");

	sprintf(log_name, LOGFILE, name);

	fprintf(stdout, "logging to file %s\n", log_name);

	if ((log_fd = open(log_name, O_WRONLY | O_CREAT | O_APPEND, 0644)) >= 0)
		syslog(LEVEL, "logging began");

	if (log_name != NULL)
		free(log_name);
}

void syslog(int ununsed, const char *fmt,...)
{
	time_t now;
	char tmp[1024];
	char *p = malloc(1024);
	va_list ap;

	if (p == NULL)
		fatal("syslog: %m");

	va_start(ap, fmt);
	(void)vsnprintf(p, 1024, fmt, ap);
	va_end(ap);

	time(&now);
	memset((char *)&tmp, 0, sizeof(tmp));

	if (!strcmp(p, last_message)) {
		if ((last_time + 60) < now) {
			snprintf(tmp, 1024, "%.15s last message repeated %d times\n", ctime(&now) + 4, times);
			write(log_fd, tmp, strlen(tmp));
			times = 0;
			time(&last_time);
		}
		times++;
		free(p);
		return;
	} else if (times > 0) {
		snprintf(tmp, 1024, "%.15s last message repeated %d times\n", ctime(&now) + 4, times);
		write(log_fd, tmp, strlen(tmp));
		times = 0;
	}
	snprintf(tmp, 1024, "%.15s %s\n", ctime(&now) + 4, p);
	write(log_fd, tmp, strlen(tmp));
	time(&last_time);
	last_message = strdup(p);

	free(p);
}
#endif
